/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004-2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sstream>
#include <boost/assign/list_of.hpp>

#include "operator_selector.h"

using namespace TkResolver;
using namespace Experimenter;
using namespace Db::Ui;




const static StringList selection_columns   = boost::assign::list_of(String("operator_id"));


OperatorSelector:: OperatorSelector(Db::Database * database,
				    Gtk::Window * parent)
  : DataSelectorInputList<OperatorInput> (database,
					  "operator",
					  selection_columns,
					  parent,
					  "Operator")
{
  const static StringList anonymous_columns   = boost::assign::list_of(String("anonymous"));
  const static StringList operator_columns = boost::assign::list_of(String("anonymous"))(String("name"))(String("email"));

  SlotListCellData * operator_column_slot =
    new SlotListCellData(sigc::mem_fun(*this, &OperatorSelector::render_operator_cell));

  const static ListColumnList display_columns = boost::assign::list_of
    (ListColumnInfo("Anonymous",   anonymous_columns,   0 ))
    (ListColumnInfo("Operator", operator_columns, operator_column_slot))
    ;

  set_display_columns(display_columns);
  
  get_view().get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
  set_selection("operator_id", database->choose_operator());
}

void
OperatorSelector:: render_operator_cell (Gtk::CellRenderer * renderer_super,
					       const Gtk::TreeModel::iterator iter)
{
  Gtk::CellRendererText * renderer = dynamic_cast<Gtk::CellRendererText *>(renderer_super);

  bool anonymous;
  iter->get_value(get_column_index("anonymous"), anonymous);

  if (anonymous)
    {
      Db::Id id;
      iter->get_value(get_column_index("operator_id"), id);
      
      std::ostringstream strm;
      strm << "Operator #";
      strm << id;
      
      renderer->property_text() = Glib::ustring(strm.str());
    }
  else
    {
      Glib::ustring name, email;
      
      iter->get_value(get_column_index("name"), name);
      iter->get_value(get_column_index("email"), email);
      
      renderer->property_text() = name + " <" + email + ">";
    }
}

Db::IdList
OperatorSelector:: get_operators()
{
  Db::IdList id_list;
  Db::Ui::Selection<Db::Id> selection(get_selection<Db::Id>());
  for (Db::Ui::Selection<Db::Id>::iterator iter(selection.begin());
       iter != selection.end();
       ++iter)
    {
      id_list.push_back((*iter)["operator_id"]);
    }
  return id_list;
}



OperatorInput:: OperatorInput(Gtk::Window & parent)
  : DataInput("Operator details", parent),
    _male("Male"),
    _anonymous("Anonymous"),
    _send_results("Send results"),
    _notes(Gtk::TextBuffer::create())
{


  set_default_size(400, 400);

  Gtk::Table * table = Gtk::manage(new Gtk::Table(8, 2));
  get_vbox()->pack_start(*table, true, true);
  
  table->attach(*Gtk::manage(new Gtk::Label("Gender",               Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 0, 1, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Date of birth",        Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 1, 2, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Anonymity",            Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 2, 3, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Results availability", Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 3, 4, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Name",                 Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 4, 5, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Place of birth",       Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 5, 6, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Email",                Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 6, 7, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Notes",                Gtk::ALIGN_LEFT|Gtk::ALIGN_TOP)), 0, 1, 7, 8, Gtk::FILL, Gtk::FILL);
  

  Gtk::HBox * gender_box = Gtk::manage(new Gtk::HBox);
  table->attach(*gender_box, 1, 2, 0, 1, Gtk::FILL, Gtk::FILL);

  gender_box->pack_start(_male, false, true);
  Gtk::RadioButton::Group gender_group(_male.get_group());
  Gtk::RadioButton * female = Gtk::manage(new Gtk::RadioButton("Female"));
  female->set_group(gender_group);

  gender_box->pack_start(*female, false, true);

  
  // set the date back 25 years
  guint year, month, day;
  _date_of_birth.get_date(year, month, day);
  _date_of_birth.select_month(month, year - 25);

  _date_of_birth.display_options(Gtk::CALENDAR_SHOW_HEADING);
  table->attach(_date_of_birth,  1, 2, 1, 2, Gtk::FILL, Gtk::FILL);
  
  _anonymous.signal_toggled().connect( sigc::mem_fun(*this, &OperatorInput::update_anonymity) );
  table->attach(_anonymous,      1, 2, 2, 3, Gtk::FILL, Gtk::FILL);

  _send_results.set_active(true);
  table->attach(_send_results,   1, 2, 3, 4, Gtk::FILL, Gtk::FILL);

  table->attach(_name,           1, 2, 4, 5, Gtk::FILL, Gtk::FILL);
  table->attach(_place_of_birth, 1, 2, 5, 6, Gtk::FILL, Gtk::FILL);
  table->attach(_email,          1, 2, 6, 7, Gtk::FILL, Gtk::FILL);

  Gtk::ScrolledWindow * notes_scroll = Gtk::manage(new Gtk::ScrolledWindow);
  notes_scroll->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
  table->attach(*notes_scroll,   1, 2, 7, 8);
  
  Gtk::TextView * notes_view = Gtk::manage(new Gtk::TextView(_notes));
  notes_scroll->add(*notes_view);
  
  update_anonymity();
  show_all_children();
}


void
OperatorInput:: update_anonymity()
{
  std::list<Gtk::Widget *> identifying_widgets =
    boost::assign::list_of
    (static_cast<Gtk::Widget *>(&_send_results))
    (static_cast<Gtk::Widget *>(&_name))
    (static_cast<Gtk::Widget *>(&_place_of_birth))
    (static_cast<Gtk::Widget *>(&_email));

  bool sensitive = !_anonymous.get_active();
  for (std::list<Gtk::Widget *>::const_iterator iter(identifying_widgets.begin());
       iter != identifying_widgets.end();
       ++iter)
    (*iter)->set_sensitive(sensitive);
}

StringList
OperatorInput:: get_insert_columns()
{
  StringList columns;
  columns.push_back(L"gender");
  columns.push_back(L"date_of_birth");
  columns.push_back(L"anonymous");
  columns.push_back(L"send_results");
  columns.push_back(L"name");
  columns.push_back(L"place_of_birth");
  columns.push_back(L"email");
  columns.push_back(L"notes");
  return columns;
}

void 
OperatorInput:: get_values_sub(Db::PStatement &stmt)
{
  stmt.setString(1, _male.get_active() ? String("M") : String("F"));

  guint year, month, day;
  _date_of_birth.get_date(year, month, day);
  stmt.setDate(2, Db::Date(year, month + 1, day));
  
  stmt.setByte(3, _anonymous.get_active());
  stmt.setByte(4, _send_results.get_active());
  stmt.setString(5, String(_name.get_text()));
  stmt.setString(6, String(_place_of_birth.get_text()));
  stmt.setString(7, String(_email.get_text()));
  stmt.setString(8, String(_notes->get_text()));
}
