/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sstream>
#include <limits>
#include <iostream>

#include "libtkcommon/debug.h"
#include "algorithm.h"

using namespace TkResolver;


Algorithm:: Algorithm (const String &spec)
  : _outcome_count(0)
{
  WDEBUG_OUT << L"Algorithm spec: '" << spec << "'" << std::endl;
  
  // extract weights
  std::wistringstream strm(spec);
  unsigned i;
  while (strm >> i)
    _weights.push_back (i);
  
  //
  // repeat the first half of the list and add it to the end
  //
  SmallNumbers reverse_list;
  if (_weights.size() > 1)
    {
      //                                         vv skip the middle weight
      for (SmallNumbers::reverse_iterator iter = ++(_weights.rbegin());
	   iter != _weights.rend();
	   ++iter)
	reverse_list.push_back(*iter);

      for (SmallNumbers::const_iterator iter = reverse_list.begin();
	   iter != reverse_list.end();
	   ++iter)
	_weights.push_back(*iter);
      
    }

  
  SmallNumbers::const_iterator weight;

  //
  // offsets
  //
  Offset offset( -(( _weights.size() - 1 ) / 2) );
  for (weight = _weights.begin();
       weight != _weights.end();
       ++weight, ++offset)
    _offsets.push_back(offset);

  //
  // outcomes
  //
  SmallNumbers weight_counts(_weights);
  bool set_some;
  do
    {
      set_some = false;
      SmallNumbers::iterator offset_iter;
      SmallNumbers::iterator count_iter;
      for (offset_iter  = _offsets.begin(), count_iter = weight_counts.begin();
	   offset_iter != _offsets.end();
	   ++offset_iter, ++count_iter)
	{
	  if (*count_iter > 0)
	    {
	      _outcome_pattern.push_back(*offset_iter);
	      --(*count_iter);
	      set_some = true;
	    }
	}
    }
  while (set_some);

  _outcome_count = _outcome_pattern.size();


  //
  // probabilities
  //
  const Probability outcome_count_d(_outcome_count);
  for (weight = _weights.begin(), i = 0;
       weight != _weights.end();
       ++weight, ++i)
    _probabilities.insert( OffsetProbabilities::value_type(_offsets[i],
							   static_cast<Probability>(*weight) / outcome_count_d) );
  
  

  
  // some nice debugging output
  /*
  std::wostringstream outcomes;
  for (i = 0; i < _outcome_count; ++i)
    outcomes << L"[" << _outcome_pattern[i %_outcome_pattern.size()] << L"]";

  std::wostringstream probabilities;
  for (OffsetProbabilities::const_iterator iter = _probabilities.begin();
       iter != _probabilities.end();
       ++iter)
    probabilities << L" P(" << iter->first << L")=" << iter->second;

  WDEBUG_OUT << L"Algorithm"
	     << L"; outcome count: " << _outcome_count
	     << L"; bit descriptor: " << _bit_desc
    	     << L"; outcomes: " << outcomes.str()
    	     << L"; probabilities:" << probabilities.str()
	     << std::endl;
  */
}


Offset
Algorithm:: choose_offset(unsigned choice)
{
  return _outcome_pattern[choice % _outcome_pattern.size()];
}

const SmallNumbers &
Algorithm:: get_weights() const
{
  return _weights;
}

const SmallNumbers &
Algorithm:: get_offsets() const
{
  return _offsets;
}

const SmallNumbers &
Algorithm:: get_outcome_pattern() const
{
  return _outcome_pattern;
}

unsigned
Algorithm:: get_outcome_count() const
{
  return _outcome_count;
}


const OffsetProbabilities &
Algorithm:: get_probabilities() const
{
  return _probabilities;
}
