/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004-2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <time.h>

#include <sstream>

#include <odbc++/setup.h>
#include <odbc++/drivermanager.h>
#include <odbc++/databasemetadata.h>

#include "libtkcommon/mbs.h"
#include "libtkcommon/exceptions.h"
#include "libtkcommon/tkvalue.h"
#include "libtkcommon/debug.h"
#include "database.h"

using namespace TkResolver;
using namespace Db;

static const char *
get_month_str(unsigned month)
{
  switch (month)
    {
    case  1: return "JAN";
    case  2: return "FEB";
    case  3: return "MAR";
    case  4: return "APR";
    case  5: return "MAY";
    case  6: return "JUN";
    case  7: return "JUL";
    case  8: return "AUG";
    case  9: return "SEP";
    case 10: return "OCT";
    case 11: return "NOV";
    case 12: return "DEC";
    }

  throw std::domain_error("unknown month number");
}

TkResolver::String
TkResolver::Db:: date_time_string(time_t t)
{
  struct tm *ltm;
  
  ltm = localtime (&t);
  if (!ltm)
    {
      throw std::runtime_error("could not convert time to local time components");
    }
  ltm->tm_year += 1900;
  ltm->tm_mon  += 1;

  std::wostringstream strm;
  strm << ltm->tm_year
       << L"-"
       << (ltm->tm_mon < 10 ? "0" : "")
       << ltm->tm_mon
       << L"-"
       << (ltm->tm_mday < 10 ? "0" : "")
       << ltm->tm_mday
       << L" "
       << ltm->tm_hour
       << L":"
       << ltm->tm_min
       << L":"
       << ltm->tm_sec;

  return strm.str();
}


/*Error:: Error (mysqlpp::Connection &connection)
  : std::runtime_error(connection.error())
{
}*/


Database:: Database (const TkResolver::String &data_source_name,
		     const TkResolver::String &username,
		     const TkResolver::String &password)
  : _connection(odbc::DriverManager::getConnection(data_source_name,
						   username,
						   password)
		)
{
  DEBUG_OUT << L"Getting database connection metadata..." << std::endl;
  odbc::DatabaseMetaData *metadata( _connection->getMetaData() );
  DEBUG_OUT << L"...got database connection metadata" << std::endl;
  
  std::wcout << L"Database product: "
	     << String(metadata->getDatabaseProductName())
	     << L"; version: "
	     << String(metadata->getDatabaseProductVersion())
	     << std::endl
	     << L"Driver: "
	     << String(metadata->getDriverName())
	     << L"; version: "
	     << String(metadata->getDriverVersion())
	     << std::endl;
  
  DEBUG_OUT << L"Initialised Database" << std::endl;
}

Database:: ~Database()
{
}

Statement *
Database:: new_statement()
{
  return _connection->createStatement(odbc::ResultSet::TYPE_FORWARD_ONLY,
				      odbc::ResultSet::CONCUR_READ_ONLY);
}

void
Database:: execute_statement(const TkResolver::String &sql)
{
  StatementPtr stmt( new_statement() );
  stmt->execute(sql);
}

DataSet *
Database:: execute_query(const TkResolver::String &sql)
{
  return new DataSet( new_statement()->executeQuery(sql) );
}

PStatement *
Database:: prepare_statement(const TkResolver::String &sql)
{
  return _connection->prepareStatement(sql,
				       odbc::ResultSet::TYPE_FORWARD_ONLY,
				       odbc::ResultSet::CONCUR_READ_ONLY);
}

void
Database:: get_table_info(const String &name)
{
  WDEBUG_OUT << L"Getting info for table " << name << L"..." << std::endl;
  
  
  WDEBUG_OUT << L"Getting metadata..." << std::endl;
  odbc::DatabaseMetaData *metadata( _connection->getMetaData() );
  WDEBUG_OUT << L"...got metadata" << std::endl;
  
  
  WDEBUG_OUT << L"Getting columns..." << std::endl;
  std::auto_ptr<odbc::ResultSet> result( metadata->getColumns(String(), String(), name, String()) );
  WDEBUG_OUT << L"...got columns" << std::endl;

  FieldInfo fields;
  WDEBUG_OUT << L"Iterating over column metadata..." << std::endl;
  for (result->next(); !result->isAfterLast(); result->next())
    {
      WDEBUG_OUT << L"  column: " << String(result->getString(4)) << std::endl;
      fields.insert(FieldInfo::value_type(result->getString(4), static_cast<odbc::Types::SQLType>(result->getShort(5)) ));
    }
  WDEBUG_OUT << L"...iterated over column metadata" << std::endl;
  
  _table_info.insert(TableInfo::value_type(name, fields));
}

void
Database:: check_table_info(const String &name)
{
  if (_table_info.find(name) == _table_info.end())
    get_table_info(name);
}

Type
Database:: get_column_type(const String &table, const String &column)
{
  WDEBUG_OUT << L"Getting type for column " << column << L" of table " << table << L"..." << std::endl;
  check_table_info(table);

  const FieldInfo &fields(_table_info[table]);
  
  FieldInfo::const_iterator iter = fields.find(column);
  if (iter == fields.end())
    throw std::runtime_error(std::string("could not find column ") +
			     static_cast<std::string>(table) + "." +
			     static_cast<std::string>(column));
  
  WDEBUG_OUT << L"...got type (" << iter->second << L")" << std::endl;
  return iter->second;
}

DataSet *
Database:: get_table_data(const String &table, const StringList &columns)
{
  std::wostringstream query;
  query << L"select ";
  StringList::const_iterator iter, next;
  for (iter = columns.begin();
       iter != columns.end();
       iter = next)
    {
      next = iter;
      ++next;
      
      query << *iter;
      if (next != columns.end())
	query << L",";
    }
  query << L" from "
	<< table
	<< L" t";
  
  String sql(query.str());
  Statement *stmt(new_statement());
  DataSet *result( new DataSet(stmt->executeQuery(sql)) );
  
  return result;
}

PStatement *
Database:: get_table_insert_statement(const String &table, const StringList &columns)
{
  std::wostringstream query;
  std::wostringstream values;

  query << L"insert into " << table << L"(";
  values << L") values (";

  StringList::const_iterator iter, next;
  for(iter = columns.begin();
      iter != columns.end();
      iter = next)
    {
      next = iter; ++next;
      
      query << *iter;
      values << L"?";
      
      if(next != columns.end())
	{
	  query << L", ";
	  values << L", ";
	}
    }
  query << values.str();
  query << L")";
  
  return prepare_statement(query.str());
}

Id
Database:: get_last_insert_id()
{
  const static String sql("select last_insert_id() from dual");
  Statement *stmt( new_statement() );
  DataSetPtr result( new DataSet(stmt->executeQuery(sql)) );
  
  bool valid( (*result)->next() );
  if (!valid)
    throw std::runtime_error("could not get last insert id");
  
  return (*result)->getLong(1);
}

void
Database:: property_type_get_details (const String &property_name,
				      Id &property_type_id,
				      String &property_value_type_name,
				      String &property_value_table_name)
{
  const static String
    sql(" select    pt.property_type_id"
	" ,         pvt.name"
	" ,         pvt.value_table"
	" from      property_type pt"
	" ,         property_value_type pvt"
	" where     pt.name = ?"
	" and       pvt.property_value_type_id = pt.property_value_type_id"
	);

  PStatement *query( prepare_statement(sql) );
  
  query->setString(1, property_name);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error(std::string("unknown property type '") + static_cast<std::string>(property_name) + "'");
  
  property_type_id = result->getLong(1);
  property_value_type_name = result->getString(2);
  property_value_table_name = result->getString(3);
}

Id
Database:: insert_property(const Id &property_type_id)
{
  StringList columns;
  columns.push_back("property_type_id");
  
  PStatementPtr stmt( get_table_insert_statement("property", columns) );
  
  stmt->setLong(1, property_type_id);
  stmt->execute();
  
  return get_last_insert_id();
}

void
Database:: insert_property_value_data (const Id &property_id,
				       const String &type_name,
				       const String &table_name,
				       const void *value)
{
  StringList columns;
  columns.push_back("property_id");
  columns.push_back("property_value");

  PStatementPtr stmt( get_table_insert_statement(table_name, columns) );
  
  stmt->setLong(1, property_id);
  
  if ("bit" == type_name)
    {
      const Bool *v(static_cast<const Bool *>(value));
      if (v->is_null())
	stmt->setNull(2, odbc::Types::BIT);
      else
	{
	  stmt->setByte(2, v->get_value());
	}
    }
  else if ("int" == type_name)
    {
      const Int *v(static_cast<const Int *>(value));
      if (v->is_null())
	stmt->setNull(2, odbc::Types::INTEGER);
      else
	stmt->setInt(2, *v);
    }
  else if ("uint" == type_name)
    {
      const Unsigned *v(static_cast<const Unsigned *>(value));
      if (v->is_null())
	stmt->setNull(2, odbc::Types::BIGINT);
      else
	stmt->setLong(2, v->get_value());
    }
  else if ("double" == type_name || "normal" == type_name)
    {
      const Double *v(static_cast<const Double *>(value));
      
      int type;
      if ("double" == type_name)
	type = odbc::Types::DOUBLE;
      else
	type = odbc::Types::DECIMAL;
      
      if (v->is_null())
	stmt->setNull(2, type);
      else
	stmt->setDouble(2, *v);
    }
  else if ("string" == type_name || "text" == type_name)
    {
      const StringValue *v(static_cast<const StringValue *>(value));
      
      int type;
      if ("string" == type_name)
	type = odbc::Types::VARCHAR;
      else
	type = odbc::Types::LONGVARCHAR;
      
      if (v->is_null())
	stmt->setNull(2, type);
      else
	stmt->setString(2, v->get_value());
    }
  else
    throw std::runtime_error(std::string("unknown property type '") + static_cast<std::string>(type_name) + "'");
  
  stmt->execute();
}

void
Database:: insert_experiment_property(const Id &experiment_id,
				      const Id &property_id)
{
  StringList columns;
  columns.push_back("experiment_id");
  columns.push_back("property_id");
  
  PStatementPtr stmt( get_table_insert_statement("experiment_property", columns) );
  
  stmt->setLong(1, experiment_id);
  stmt->setLong(2, property_id);
  stmt->execute();
}

Id
Database:: experiment_set_property(const Id &experiment_id,
				   const String &property_name,
				   const void *value)
{
  WDEBUG_OUT << L"Setting experiment property '" << property_name << "'.." << std::endl;
  Id property_type_id;
  String property_value_type_name;
  String property_value_table_name;
  
  WDEBUG_OUT << L"Getting type details..." << std::endl;
  property_type_get_details (property_name,
			     property_type_id,
			     property_value_type_name,
			     property_value_table_name);
  WDEBUG_OUT << L"...got type details (type name: '"
	     << property_value_type_name
	     << L"'; table name: '"
	     << property_value_table_name
	     << L"')"
	     << std::endl;
  
  WDEBUG_OUT << L"Inserting property..." << std::endl;
  Id property_id( insert_property(property_type_id) );
  WDEBUG_OUT << L"...inserted property (id: " << property_id << L")" << std::endl;
  
  WDEBUG_OUT << L"Inserting experiment property..." << std::endl;
  insert_experiment_property(experiment_id, property_id);
  WDEBUG_OUT << L"...inserted experiment property" << std::endl;
  
  WDEBUG_OUT << L"Inserting property value data..." << std::endl;
  insert_property_value_data(property_id,
			     property_value_type_name,
			     property_value_table_name,
			     value);
  WDEBUG_OUT << L"...inserted property value data" << std::endl;

  WDEBUG_OUT << L"..set experiment property '" << property_name << L"'" << std::endl;
  return property_id;
}


Id
Database:: lookup_location(const String &name)
{
  const static String sql("select location_id from location where name = ?");
  PStatement *stmt( prepare_statement(sql) );
  stmt->setString(1, name);
  
  DataSet result( stmt->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error(std::string("unknown location '") + static_cast<std::string>(name) + "'");
  
  return result->getLong(1);
}

Id
Database:: insert_experiment(bool test,
			     const Db::DateTime &date_time,
			     const Id &location_id,
			     const Random::Bits &randomness)
{
  StringList columns;
  columns.push_back("test");
  columns.push_back("date_time");
  columns.push_back("location_id");
  columns.push_back("randomness_bits");
  columns.push_back("randomness");
  
  PStatementPtr stmt( get_table_insert_statement("experiment", columns) );
  
  stmt->setBoolean(1, test);
  stmt->setTimestamp(2, DateTime(date_time));
  stmt->setLong(3, location_id);
  stmt->setLong(4, randomness.size());
  
  Random::BitIStream *randomness_stream( randomness.get_data() );
  stmt->setBinaryStream(5, randomness_stream, randomness.get_data_len());
  
  
  stmt->execute();
  
  delete randomness_stream;
  
  return get_last_insert_id();
}

void
Database:: experiment_insert_operators(const Id &experiment_id,
				       const IdList &operators)
{
  StringList columns;
  columns.push_back("experiment_id");
  columns.push_back("operator_id");
  
  PStatementPtr stmt( get_table_insert_statement("experiment_operator", columns) );
  
  stmt->setLong(1, experiment_id);
  
  for (IdList::const_iterator iter(operators.begin());
       iter != operators.end();
       ++iter)
    {
      stmt->setLong(2, *iter);
      stmt->execute();
    }
}

void
Database:: experiment_insert_result(const Id &experiment_id,
				    const Result &result)
{
  StringList columns;
  columns.push_back("experiment_id");
  columns.push_back("line_index");
  columns.push_back("offset_index");
  columns.push_back("offset");
  
  PStatementPtr stmt( get_table_insert_statement("experiment_result", columns) );
  
  stmt->setLong(1, experiment_id);
  
  unsigned ntraces( result.begin()->size() );
  unsigned line_index(0);
  unsigned offset_index;
  for (Result::const_iterator iter(result.begin());
       iter != result.end();
       ++iter, ++line_index)
    {
      stmt->setInt(2, line_index);
      
      for (offset_index = 0; offset_index < ntraces; ++offset_index)
	{
	  stmt->setInt(3, offset_index);
	  stmt->setInt(4, (*iter)[offset_index]);
	  
	  stmt->execute();
	}
    }
}

Id
Database:: insert_experiment(bool test,
			     const Db::DateTime &date_time,
			     const Id &location_id,
			     const Random::Bits &randomness,
			     
			     const StringValue &algorithm_spec,
			     const Unsigned &bit_multiplier,
			     const Unsigned &trace_count,
			     const Unsigned &time,
			     const Bool &scrolled,
			     const Double &speed,
			     const Double &horizon,
			     const StringValue &notes,
			     const StringValue &intent_code,
			     const Double &ability_rating,
			     const Unsigned &width,
			     const Unsigned &height,
			     const String &retriever,
			     const Bool &control,
			     
			     const IdList &operators,
			     const Result &result)
{
  DEBUG_OUT << "Inserting experiment.." << std::endl;
  
  DEBUG_OUT << "Inserting experiment table..." << std::endl;
  Id id( insert_experiment(test, date_time, location_id, randomness) );
  DEBUG_OUT << "...inserted experiment table" << std::endl;
  
  DEBUG_OUT << "Setting properties..." << std::endl;
  experiment_set_property(id, "algorithm spec", &algorithm_spec);
  experiment_set_property(id, "bit multiplier", &bit_multiplier);
  experiment_set_property(id, "trace count", &trace_count);
  experiment_set_property(id, "time", &time);
  experiment_set_property(id, "scrolled", &scrolled);
  experiment_set_property(id, "speed", &speed);
  experiment_set_property(id, "horizon", &horizon);
  experiment_set_property(id, "notes", &notes);
  experiment_set_property(id, "intent code", &intent_code);
  experiment_set_property(id, "ability rating", &ability_rating);
  experiment_set_property(id, "screen width", &width);
  experiment_set_property(id, "screen height", &height);
  StringValue retriever_value(retriever);
  experiment_set_property(id, "retriever name", &retriever_value);
  experiment_set_property(id, "control", &control);
  DEBUG_OUT << "...set properties" << std::endl;
  
  
  DEBUG_OUT << "Inserting operators..." << std::endl;
  experiment_insert_operators(id, operators);
  DEBUG_OUT << "...inserted operators" << std::endl;

  DEBUG_OUT << "Inserting result..." << std::endl;
  experiment_insert_result(id, result);
  DEBUG_OUT << "...inserted result" << std::endl;
  
  DEBUG_OUT << "..inserted experiment (id: " << id << ")" << std::endl;
  return id;
}


Id
Database:: choose_location()
{
  String
    sql(" select   e.location_id"
	" from     experiment e"
	" order by e.date_time desc"
	);
  DataSetPtr result( execute_query(sql) );
  
  bool valid( (*result)->next() );
  if (!valid)
    {
      sql = "select location_id from location";
      result.reset( execute_query(sql) );
      
      valid = (*result)->next();
    }
  
  if (!valid)
    throw std::runtime_error("could not choose location");

  return (*result)->getLong(1);
}

Id
Database:: choose_operator()
{
  String
    sql(" select   eo.operator_id"
	" from     experiment e"
	" ,        experiment_operator eo"
	" where    e.experiment_id = eo.experiment_id"
	" order by e.date_time desc"
	);
  DataSetPtr result( execute_query(sql) );
  
  bool valid( (*result)->next() );
  if (!valid)
    {
      sql = "select operator_id from operator";
      result.reset( execute_query(sql) );
      
      valid = (*result)->next();
    }
  
  if (!valid)
    throw std::runtime_error("could not choose operator");

  return (*result)->getLong(1);
}

Code
Database:: choose_intent(unsigned trace_count)
{
  const static String
    sql(" select   i.intent_code"
	" from     intent_domain id"
	" ,        intent i"
	" where    id.trace_count = ?"
	" and      i.intent_domain_id = id.intent_domain_id"
	" order by ("
	"            select   count(*)"
	"            from     property_value_string pvs"
	"            ,        property p"
	"            ,        property_type pt"
	"            ,        experiment_property ep"
	"            ,        experiment e"
	"            where    pvs.property_value        = i.intent_code"
	"            and      p.property_id             = pvs.property_id"
	"            and      pt.property_type_id       = p.property_type_id"
	"            and      pt.name                   = ?"
	"            and      ep.property_id            = p.property_id"
	"            and      e.experiment_id           = ep.experiment_id"
	"	   )"
	);
  
  PStatement* query( prepare_statement(sql) );
  
  query->setInt(1, static_cast<int>(trace_count));
  query->setString(2, String("intent code"));
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error("could not choose intent");
  
  return result->getString(1);
}

IntentInfoList
Database:: get_intent_options(unsigned trace_count)
{
  const static String
    sql(" select   i.intent_code"
	" ,        i.description"
	" from     intent_domain id"
	" ,        intent i"
	" where    id.trace_count = ?"
	" and      i.intent_domain_id = id.intent_domain_id"
	);
  PStatement *query( prepare_statement(sql) );
  
  query->setInt(1, trace_count);
  
  DataSet result( query->executeQuery() );
  
  IntentInfoList info;
  for (result->next(); !result->isAfterLast(); result->next())
    info.push_back( IntentInfo(result->getString(1), result->getString(2)) );
  
  return info;
}

IdList
Database:: get_experiments(const TkResolver::String &property_expression)
{
  // TODO: write query language
  return IdList();
}

DataSet *
Database:: get_results(const Id &experiment_id)
{
  const static String
    sql(" select   offset"
	" from     experiment_result"
	" where    experiment_id =  ?"
	" order by line_index"
	" ,        offset_index"
	);
  
  PStatement *stmt( prepare_statement(sql) );
  stmt->setLong(1, experiment_id);
  
  return new DataSet(stmt->executeQuery());
}

unsigned
Database:: get_experiment_screen_width(const Id &experiment_id)
{
  const static String
    sql(" select   pvu.property_value"
	" from     property_value_uint pvu"
	" ,        property p"
	" ,        property_type pt"
	" ,        experiment_property ep"
	" ,        experiment e"
	" where    p.property_id             = pvu.property_id"
	" and      pt.property_type_id       = p.property_type_id"
	" and      pt.name                   = ?"
	" and      ep.property_id            = p.property_id"
	" and      e.experiment_id           = ep.experiment_id"
	" and      e.experiment_id           = ?"
	);
					 
  PStatementPtr query( prepare_statement(sql) );
  
  query->setString(1, "screen width");
  query->setLong(2, experiment_id);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error("could not find screen width for experiment");
  
  return result->getLong(1);
}
